#include "layout.h"

int desktop_remove(Dsktp* dsk){
  if(!dsk)return 1;
  Monitor* m=dsk->mon;
  if(!m)return 1;
  if(dsk->next){
    dsk->next->prev=dsk->prev;
  }else{
    m->lastdesktop=dsk->prev;
  }
  if(dsk->prev){
    dsk->prev->next=dsk->next;
  }else{
    m->firstdesktop=dsk->next;
  }
  //fix indices
  for(Dsktp* d=dsk->next;d;d=d->next){
    d->id--;
  }

  dsk->next=0;
  dsk->prev=0;
  m->desktopcount--;
  dsk->mon=0;
  dsk->id=-1;
  return 0;
}

int desktop_add(Monitor* mon, Dsktp* dsk){
  if(!mon || !dsk)return 1;
  if(dsk->mon)return 1;
  dsk->next=0;
  dsk->prev=mon->lastdesktop;
  if(dsk->prev)dsk->prev->next=dsk;
  mon->lastdesktop=dsk;
  if(!mon->firstdesktop)mon->firstdesktop=dsk;
  dsk->mon=mon;
  dsk->id=mon->desktopcount;
  mon->desktopcount++;
  return 0;
}


int layer_stack_remove(Layer* lay){
  if(!lay)return 1;
  Monitor* m=lay->mon;
  if(!m)return 1;
  if(lay->snext){
    lay->snext->sprev=lay->sprev;
  }else{
    m->slastlayer=lay->sprev;
  }
  if(lay->sprev){
    lay->sprev->snext=lay->snext;
  }else{
    m->sfirstlayer=lay->snext;
  }
  lay->snext=0;
  lay->sprev=0;
  return 0;
}

int layer_stack_add_front(Monitor* mon, Layer* lay){
  if(!mon || !lay)return 1;
  lay->sprev=0;
  lay->snext=mon->sfirstlayer;
  if(lay->snext)lay->snext->sprev=lay;
  mon->sfirstlayer=lay;
  if(!mon->slastlayer)mon->slastlayer=lay;
  return 0;
}

int layer_stack_add_back(Monitor* mon, Layer* lay){
  if(!mon || !lay)return 1;
  lay->snext=0;
  lay->sprev=mon->slastlayer;
  if(lay->sprev)lay->sprev->snext=lay;
  mon->slastlayer=lay;
  if(!mon->sfirstlayer)mon->sfirstlayer=lay;
  return 0;
}

int layer_stack_to_front(Layer* lay){
  if(!lay->sprev)return 0;
  if(layer_stack_remove(lay))return 1;
  if(layer_stack_add_front(lay->mon,lay))return 1;
  return 0;
}

int layer_stack_to_back(Layer* lay){
  if(!lay->snext)return 0;
  if(layer_stack_remove(lay))return 1;
  if(layer_stack_add_back(lay->mon,lay))return 1;
  return 0;
}

int layer_remove(Layer* lay){
  if(!lay)return 1;
  if(!lay->prev){
    say("cannot remove zeroth layer\n");
    return 1;
  }
  if(lay->tilecount){
    say("cannot remove nonempty layer\n");
    return 1;
  }
  if(lay->snext || lay->sprev){
    say("cannot remove layer if it is still on stack\n");
    return 1;
  }
  Monitor* m=lay->mon;
  if(!m)return 1;
  if(lay->next){
    lay->next->prev=lay->prev;
  }else{
    m->lastlayer=lay->prev;
  }
  if(lay->prev){
    lay->prev->next=lay->next;
  }else{
    m->firstlayer=lay->next;
  }
  lay->next=0;
  lay->prev=0;
  m->layercount--;
  lay->mon=0;
  return 0;
}

int layer_add(Monitor* mon, Layer* lay){
  if(!mon || !lay)return 1;
  if(lay->mon)return 1;
  lay->next=0;
  lay->prev=mon->lastlayer;
  if(lay->prev)lay->prev->next=lay;
  mon->lastlayer=lay;
  if(!mon->firstlayer)mon->firstlayer=lay;
  lay->mon=mon;
  mon->layercount++;
  return 0;
}


int tile_remove(Tile* t){
  if(!t)return 1;
  if(t->clientcount){
    say("cannot remove nonempty tile\n");
    return 1;
  }
  Layer* l=t->layer;
  if(!l)return 1;
  if(t->next){
    t->next->prev=t->prev;
  }else{
    l->lasttile=t->prev;
  }
  if(t->prev){
    t->prev->next=t->next;
  }else{
    l->firsttile=t->next;
  }
  t->next=0;
  t->prev=0;
  l->tilecount--;
  t->layer=0;

  //remove rules that link to us
  Monitor* m=l->mon;
  if(!m)return 1;
  if(!m->rulz)return 0;
  if(m->rulz->target==t){
    Rule* todelete=m->rulz;
    m->rulz=m->rulz->next;
    destroy_rule(todelete);
  }
  for(Rule* r=m->rulz;r->next;r=r->next){
    if(r->next->target==t){
      Rule* todelete=r->next;
      r->next=r->next->next;
      destroy_rule(todelete);
    }
  }
  //
  return 0;
}

int tile_add(Layer* lay, Tile* t){
  if(!lay || !t)return 1;
  if(t->layer)return 1;
  t->next=0;
  t->prev=lay->lasttile;
  if(t->prev)t->prev->next=t;
  lay->lasttile=t;
  if(!lay->firsttile)lay->firsttile=t;
  t->layer=lay;
  lay->tilecount++;
  tile_recalc(t);
  //say("tile: %s %d %d %d %d\n",t->name,t->x,t->y,t->w,t->h);
  return 0;
}

int tile_recalc(Tile* t){
  if(!t)return 1;
  if(!t->layer)return 1;
  Monitor* m=t->layer->mon;
  if(!m)return 1;
  t->x=t->xa+(int)(t->xr*m->mw);
  t->y=t->ya+(int)(t->yr*m->mh);
  t->w=t->wa+(int)(t->wr*m->mw);
  t->h=t->ha+(int)(t->hr*m->mh);
  return 0;
}

int client_stack_remove(Client* c){
  if(!c)return 1;
  Tile* t=c->tile;
  Layer* l=c->layer;
  if(!t)return 1;
  if(!l)return 1;
  if(c->snext){
    c->snext->sprev=c->sprev;
  }else{
    t->lastclient=c->sprev;
  }
  if(c->sprev){
    c->sprev->snext=c->snext;
  }else{
    t->firstclient=c->snext;
  }
  c->snext=0;
  c->sprev=0;
  l->clientcount--;
  t->clientcount--;
  c->tile=0;
  c->layer=0;
  return 0;
}


int client_stack_add_front(Tile* t, Client* c){
  if(!t || !c)return 1;
  if(c->tile)client_stack_remove(c);
  Layer* l=t->layer;
  c->sprev=0;
  c->snext=t->firstclient;
  if(c->snext)c->snext->sprev=c;
  t->firstclient=c;
  if(!t->lastclient)t->lastclient=c;
  t->clientcount++;
  l->clientcount++;
  c->tile=t;
  c->layer=l;
  return 0;
}

int client_stack_add_back(Tile* t, Client* c){
  if(!t || !c)return 1;
  Layer* l=t->layer;
  c->snext=0;
  c->sprev=t->lastclient;
  if(c->sprev)c->sprev->snext=c;
  t->lastclient=c;
  if(!t->firstclient)t->firstclient=c;
  t->clientcount++;
  l->clientcount++;
  c->tile=t;
  c->layer=l;
  return 0;
}

int client_stack_add_to_layer(Layer* lay, Client* c, int level){
  switch(level){
    case AddlvlSticky:
      if(!lay->visible)return 1;
      for(Tile* t=lay->firsttile;t;t=t->next){
	if(t->sticky==StickStick && tile_is_free(t)){
	  client_stack_add_front(t,c);
	  return 0;
	}
      }
      break;
    case AddlvlEmpty:
      if(!lay->visible)return 1;
      for(Tile* t=lay->firsttile;t;t=t->next){
	if(t->sticky!=StickSlippery && t->sticky!=StickAvoid && t->clientcount==0){
	  client_stack_add_front(t,c);
	  return 0;
	}
      }
      break;
    case AddlvlFree:
      if(!lay->visible)return 1;
      Tile* best=0;
      int min=65536;//if you have more windows it doesn't matter anyway
      for(Tile* t=lay->firsttile;t;t=t->next){
	if(t->sticky!=StickSlippery && t->sticky!=StickAvoid && tile_is_free(t)){
	  if(t->clientcount<min){
	    min=t->clientcount;
	    best=t;
	  }
	}
      }
      if(best){
	//say("foundbest, %s\n",best->name);
	client_stack_add_front(best,c);
	return 0;
      }
      break;
    case AddlvlOther:
      for(Tile* t=lay->firsttile;t;t=t->next){
	if(t->sticky!=StickAvoid && tile_is_free(t)){
	  client_stack_add_front(t,c);
	  return 0;
	}
      }
      break;

  }
  return 1;
}

int client_stack_apply_rules(Monitor* mon, Client* c){
  for(Rule* r=mon->rulz;r;r=r->next){
    if((!r->target) || !tile_is_free(r->target) || (r->visible_only && !r->target->layer->visible))continue;
    int passed=1;
    if(r->name){
      passed=passed&&(!strcmp(r->name,c->name));
    }
    if(r->role){
      passed=passed&&(!strcmp(r->role,c->role));
    }
    if(r->classN){
      passed=passed&&(!strcmp(r->classN,c->classN));
    }
    if(r->classC){
      passed=passed&&(!strcmp(r->classC,c->classC));
    }
    if(passed){
      client_stack_add_front(r->target,c);
      return 0;
    }
  }
  return 1;
}

int client_stack_add_to_monitor(Monitor* mon, Client* c){
  /*
  for(Rule* r=mon->rulz;r;r=r->next){
    if((!r->target) || !tile_is_free(r->target) || (r->visible_only && !r->target->layer->visible))continue;
    int passed=1;
    if(r->name){
      passed=passed&&(!strcmp(r->name,c->name));
    }
    if(r->role){
      passed=passed&&(!strcmp(r->role,c->role));
    }
    if(r->classN){
      passed=passed&&(!strcmp(r->classN,c->classN));
    }
    if(r->classC){
      passed=passed&&(!strcmp(r->classC,c->classC));
    }
    if(passed){
      if(r->x!=RULE_INTNONE)c->x=r->target->x+r->x+r->xr*r->target->w;
      if(r->y!=RULE_INTNONE)c->y=r->target->y+r->y+r->yr*r->target->h;
      if(r->w!=RULE_INTNONE)c->w=r->w+r->wr*r->target->w;
      if(r->h!=RULE_INTNONE)c->h=r->h+r->hr*r->target->h;
      if(r->float_state!=RULE_INTNONE)c->float_state=r->float_state;
      if(r->stack_level!=RULE_INTNONE)c->stacktype=r->stack_level;
      client_stack_add_front(r->target,c);
      return 0;
    }
  }
  */

  for(int lvl=AddlvlSticky;lvl<=AddlvlOther;lvl++){
    for(Layer* l=mon->sfirstlayer;l;l=l->snext){
      if(!l->prev)continue;
      //say("adding 0x%x to level %d layer %s\n",c->win,lvl,l->name);
      if(!client_stack_add_to_layer(l,c,lvl))return 0;
    }
  }
  client_stack_add_front(mon->firstlayer->firsttile,c);

  return 1;
}

Rule* client_find_rule(Monitor* mon, Client* c){
  for(Rule* r=mon->rulz;r;r=r->next){
    int passed=1;
    if(r->name){
      passed=passed&&(!strcmp(r->name,c->name));
    }
    if(r->role){
      passed=passed&&(!strcmp(r->role,c->role));
    }
    if(r->classN){
      passed=passed&&(!strcmp(r->classN,c->classN));
    }
    if(r->classC){
      passed=passed&&(!strcmp(r->classC,c->classC));
    }
    if(passed){
      return r;
    }
  }
  return 0;
}


int client_post_apply_rules(Monitor* mon, Client* c){
  for(Rule* r=mon->rulz;r;r=r->next){
    if((r->target) && (r->target != c->tile))continue;
    int passed=1;
    if(r->name){
      passed=passed&&(!strcmp(r->name,c->name));
    }
    if(r->role){
      passed=passed&&(!strcmp(r->role,c->role));
    }
    if(r->classN){
      passed=passed&&(!strcmp(r->classN,c->classN));
    }
    if(r->classC){
      passed=passed&&(!strcmp(r->classC,c->classC));
    }
    if(passed){
      if(r->x!=RULE_INTNONE)c->x=c->tile->x+r->x+r->xr*c->tile->w;
      if(r->y!=RULE_INTNONE)c->y=c->tile->y+r->y+r->yr*c->tile->h;
      if(r->w!=RULE_INTNONE)c->w=r->w+r->wr*c->tile->w;
      if(r->h!=RULE_INTNONE)c->h=r->h+r->hr*c->tile->h;
      if(r->float_state!=RULE_INTNONE)c->float_state=r->float_state;
      if(r->stack_level!=RULE_INTNONE)c->stacktype=r->stack_level;
      client_reshape(c,0);
      return 0;
    }
  }
  return 1;
}

int client_move_to_tile(Tile* t, Client* c){
  if(c->tile == t)return 0;//no need to do anything
  if(client_stack_remove(c))return 1;
  if(client_stack_add_front(t,c))return 1;
  return 0;
}

int client_focus_remove(Client* c){
  if(!c)return 1;
  Monitor* m=c->mon;
  if(!m)return 1;
  if(c->fnext){
    c->fnext->fprev=c->fprev;
  }else{
    m->lastfocus=c->fprev;
  }
  if(c->fprev){
    c->fprev->fnext=c->fnext;
  }else{
    m->firstfocus=c->fnext;
  }
  c->fnext=0;
  c->fprev=0;
  return 0;
}

int client_focus_add_front(Client* c){
  if(!c)return 1;
  Monitor* m=c->mon;
  if(!m)return 1;
  c->fprev=0;
  c->fnext=m->firstfocus;
  if(c->fnext)c->fnext->fprev=c;
  m->firstfocus=c;
  if(!m->lastfocus)m->lastfocus=c;
  return 0;
}

int client_focus_add_second(Client* c){
  if(!c)return 1;
  Monitor* m=c->mon;
  if(!m)return 1;
  if(!m->firstfocus){
    m->firstfocus=c;
    c->fprev=0;
    c->fnext=0;
  }else{
    c->fnext=m->firstfocus->fnext;
    if(c->fnext){
      c->fnext->fprev=c;
    }else{
      m->lastfocus=c;
    }
    m->firstfocus->fnext=c;
    c->fprev=m->firstfocus;
  }
  return 0;
}


int client_focus_add_back(Client* c){
  if(!c)return 1;
  Monitor* m=c->mon;
  if(!m)return 1;
  c->fnext=0;
  c->fprev=m->lastfocus;
  if(c->fprev)c->fprev->fnext=c;
  m->lastfocus=c;
  if(!m->firstfocus)m->firstfocus=c;
  return 0;
}

int client_focus_to_front(Client* c){
  client_focus_remove(c);
  client_focus_add_front(c);
  return 0;
}

int client_focus_to_back(Client* c){
  client_focus_remove(c);
  client_focus_add_back(c);
  return 0;
}

int client_addtransient(Client* parent, Client* trans){
  if(!parent)return 1;
  if(!trans)return 1;
  if(parent->children){
    Client* c=parent->children;
    while(c->chld_next)c=c->chld_next;
    c->chld_next=trans;
    trans->chld_next=0;
  }else{
    parent->children=trans;
    trans->chld_next=0;
  }
  trans->owner=parent;
  return 0;
}

int client_remtransient(Client* trans){
  if(!trans)return 1;
  Client* parent=trans->owner;
  if(!parent)return 1;
  if(parent->children==trans){
    parent->children=trans->chld_next;
  }else{
    Client* c=parent->children;
    while(c->chld_next && c->chld_next!=trans)c=c->chld_next;
    if(c->chld_next!=trans){
      say("didn't find myself in my father!\n");
      return 1;
    } else {
      c->chld_next=trans->chld_next;
    }
  }
  trans->chld_next=0;
  trans->owner=0;
  return 0;
}

int client_remove(Client* c){
  if(!c)return 1;
  if(c->snext || c->sprev){
    say("cannot remove client that is still on stack\n");
    return 1;
  }
  Monitor* m=c->mon;
  if(!m)return 1;
  if(c->next){
    c->next->prev=c->prev;
  }else{
    m->lastclient=c->prev;
  }
  if(c->prev){
    c->prev->next=c->next;
  }else{
    m->firstclient=c->next;
  }
  c->next=0;
  c->prev=0;
  m->clientcount--;
  c->mon=0;
  return 0;
}

int client_add(Monitor* mon, Client* c){
  if(!mon || !c)return 1;
  if(c->mon)return 1;
  c->prev=0;
  c->next=mon->firstclient;
  if(c->next)c->next->prev=c;
  mon->firstclient=c;
  if(!mon->lastclient)mon->lastclient=c;
  c->mon=mon;
  mon->clientcount++;
  return 0;
}

void client_keepintile(Tile* t, int* x, int* y){
  if(*x<t->x+SNAPWID)*x=t->x;
  else if(*x>t->x+t->w-SNAPWID)*x=t->x+t->w;
  if(*y<t->y+SNAPWID)*y=t->y;
  else if(*y>t->y+t->h-SNAPWID)*y=t->y+t->h;
}


int client_movecorner(Client* c, int leftright, int topbottom){
  if(!c)return 1;
  Tile* t=c->tile;
  if(!t)return 1;

  //fixed vertex
  int xf=!leftright?c->x:c->x+c->w;
  int yf=!topbottom?c->y:c->y+c->h;

  //moving vertex
  int xx=leftright?c->x:c->x+c->w;
  int yy=topbottom?c->y:c->y+c->h;
  client_keepintile(t,&xx,&yy);
  //new tile-compliant suggestion
  if(leftright){
    c->x=xx;
    c->w=xf-xx;
  }else{
    c->x=xf;
    c->w=xx-xf;
  }
  if(topbottom){
    c->y=yy;
    c->h=yf-yy;
  }else{
    c->y=yf;
    c->h=yy-yf;
  }
  //our "w" and "h" include border
  c->w-=2*BWID;
  c->h-=2*BWID;

  //correct the w,h according to wishes

  //obey the increments and minsize
  int baseismin = c->basew == c->minw && c->baseh == c->minh;
  if(!baseismin) { /* temporarily remove base dimensions */
    c->w -= c->basew;
    c->h -= c->baseh;
  }
  /* adjust for aspect limits */
  if(c->mina > 0 && c->maxa > 0) {
    if(c->maxa < (float)c->w /c->h){
      c->w = c->h * c->maxa + 0.5;
    }else if(c->mina < (float)c->h /c->w){
      c->h = c->w * c->mina + 0.5;
    }
  }
  if(baseismin) { /* increment calculation requires this */
    c->w -= c->basew;
    c->h -= c->baseh;
  }
  /* adjust for increment value */
  if(c->incw)
    c->w -= c->w % c->incw;
  if(c->inch)
    c->h -= c->h % c->inch;
  /* restore base dimensions */
  c->w += c->basew;
  c->h += c->baseh;
  if(c->w<c->minw)c->w = c->minw;
  if(c->h<c->minh)c->h = c->minh;
  if(c->maxw && c->w>c->maxw)
    c->w = c->maxw;
  if(c->maxh && c->h>c->maxh)
    c->h = c->maxh;

  //our "w" and "h" include border
  c->w+=2*BWID;
  c->h+=2*BWID;

  //now move the fixed corner back where it belongs
  if(leftright){
    c->x=xf-c->w;
  }
  if(topbottom){
    c->y=yf-c->h;
  }

  return 0;
}

static void dosnap(Client* c){
  Tile* t=c->tile;

  int xsnap=SNAPWID+1;
  int ysnap=SNAPWID+1;

  int newxs, newys;

  //leftsnap
  newxs=t->x-c->x;
  if(abs(newxs)<abs(xsnap))xsnap=newxs;
  //rightsnap
  newxs=(t->x+t->w)-(c->x+c->w);
  if(abs(newxs)<abs(xsnap))xsnap=newxs;
  //tiles
  for(Client* c2=t->firstclient;c2;c2=c2->snext){
    if(c2==c)continue;
    if(c2->float_state!=FloatFloat)continue;
    //no vertical overlap
    if(! (c->y+c->h>c2->y && c->y<c2->y+c2->h))continue;
    //left-left
    newxs=c2->x-c->x;
    if(abs(newxs)<abs(xsnap))xsnap=newxs;
    //left-right
    newxs=c2->x-(c->x+c->w);
    if(abs(newxs)<abs(xsnap))xsnap=newxs;
    //right-left
    newxs=(c2->x+c2->w)-c->x;
    if(abs(newxs)<abs(xsnap))xsnap=newxs;
    //right-right
    newxs=(c2->x+c2->w)-(c->x+c->w);
    if(abs(newxs)<abs(xsnap))xsnap=newxs;
  }


  //leftsnap
  newys=t->y-c->y;
  if(abs(newys)<abs(ysnap))ysnap=newys;
  //rightsnap
  newys=(t->y+t->h)-(c->y+c->h);
  if(abs(newys)<abs(ysnap))ysnap=newys;
  //tiles
  for(Client* c2=t->firstclient;c2;c2=c2->snext){
    if(c2==c)continue;
    if(c2->float_state!=FloatFloat)continue;
    //no horizontal overlap
    if(! (c->x+c->w>c2->x && c->x<c2->x+c2->w))continue;
    //left-left
    newys=c2->y-c->y;
    if(abs(newys)<abs(ysnap))ysnap=newys;
    //left-right
    newys=c2->y-(c->y+c->h);
    if(abs(newys)<abs(ysnap))ysnap=newys;
    //right-left
    newys=(c2->y+c2->h)-c->y;
    if(abs(newys)<abs(ysnap))ysnap=newys;
    //right-right
    newys=(c2->y+c2->h)-(c->y+c->h);
    if(abs(newys)<abs(ysnap))ysnap=newys;
  }


  if(abs(xsnap)<=SNAPWID)c->x+=xsnap;
  if(abs(ysnap)<=SNAPWID)c->y+=ysnap;
}

int client_rehint(Client* c, int snap){
  if(!c)return 1;
  Tile* t=c->tile;
  if(!t)return 1;
  //snapping
  if(!snap)dosnap(c);

  //put in the bottom-right corner
  if(c->x+c->w > t->x+t->w){
    c->x=t->x+t->w-c->w;
  }
  if(c->y+c->h > t->y+t->h){
    c->y=t->y+t->h-c->h;
  }
  //put in the top-left corner
  if(c->x<t->x){
    c->x=t->x;
  }
  if(c->y<t->y){
    c->y=t->y;
  }
  //resize if sticking out
  if(c->w>t->w){
    c->w=t->w;
  }
  if(c->h>t->h){
    c->h=t->h;
  }
  //our "w" and "h" include border
  c->w-=2*BWID;
  c->h-=2*BWID;

  //obey the increments and minsize
  int baseismin = c->basew == c->minw && c->baseh == c->minh;
  if(!baseismin) { // temporarily remove base dimensions
    c->w -= c->basew;
    c->h -= c->baseh;
  }
  // adjust for aspect limits
  if(c->mina > 0 && c->maxa > 0) {
    if(c->maxa < (float)c->w /c->h){
      c->w = c->h * c->maxa + 0.5;
    }else if(c->mina < (float)c->h /c->w){
      c->h = c->w * c->mina + 0.5;
    }
  }
  if(baseismin) { // increment calculation requires this
    c->w -= c->basew;
    c->h -= c->baseh;
  }
  // adjust for increment value
  if(c->incw)
    c->w -= c->w % c->incw;
  if(c->inch)
    c->h -= c->h % c->inch;
  // restore base dimensions
  c->w += c->basew;
  c->h += c->baseh;
  if(c->w<c->minw)c->w = c->minw;
  if(c->h<c->minh)c->h = c->minh;
  if(c->maxw && c->w>c->maxw)
    c->w = c->maxw;
  if(c->maxh && c->h>c->maxh)
    c->h = c->maxh;


  if(c->w<1)c->w=1;
  if(c->h<1)c->h=1;

  //our "w" and "h" include border
  c->w+=2*BWID;
  c->h+=2*BWID;

  return 0;
}


int client_reshape(Client* c, int snap){
  if(!c)return 1;
  Tile* t=c->tile;
  if(!t)return 1;
  client_rehint(c,snap);
  //if transient and no type, or if type set the floating flag
  int should_float=((c->owner) && !(c->hints&MY_HINT_TYPES)) || (c->hints&MY_HINT_FLOAT);
  if(c->float_state==FloatFull){//fullscreen
    c->xx=0;
    c->yy=0;
    c->ww=c->mon->mw;
    c->hh=c->mon->mh;
    c->bw=0;
    return 0;
  }
  if(c->float_state==FloatMax || (t->force_maximize && !should_float)){//maximized
    c->xx=t->x;
    c->yy=t->y;
    c->ww=t->w;
    c->hh=t->h;
    c->bw=(c->hints&MY_HINT_NOBRD)?0:BWID;
    return 0;
  }
  if(c->float_state==FloatFloat){//floating window
    c->xx=c->x;
    c->yy=c->y;
    c->ww=c->w;
    c->hh=c->h;
    c->bw=(c->hints&MY_HINT_NOBRD)?0:BWID;
    return 0;
  }
  return 1;//never happens
}

int client_move(Client* c, int x, int y, int x0, int y0, Client* oldc){
  if(!c || !oldc)return 1;
  if(!c->tile)return 1;
  c->x=oldc->x+(x-x0);
  c->y=oldc->y+(y-y0);
  c->w=oldc->w;
  c->h=oldc->h;
  for(Tile* t=oldc->layer->firsttile;t;t=t->next){
    if(!tile_is_free(t))continue;
    if(x>=t->x &&
       x<t->x+t->w &&
       y>=t->y &&
       y<t->y+t->h){
      //temporary: later you need to use oldc and remove & put back the client
      oldc->tile=t;
      break;
    }
  }
  return client_reshape(c,0);
}

int client_resize(Client* c, int x, int y, int x0, int y0, Client* oldc){
  if(!c || !oldc)return 1;

  c->x=oldc->x;
  c->y=oldc->y;
  c->w=oldc->w;
  c->h=oldc->h;

  int midX=oldc->x+oldc->w/2;
  int midY=oldc->y+oldc->h/2;

  if(x0<midX){
    c->x+=x-x0;
    c->w-=x-x0;
  }else{
    c->w+=x-x0;
  }
  if(y0<midY){
    c->y+=y-y0;
    c->h-=y-y0;
  }else{
    c->h+=y-y0;
  }

  client_movecorner(c,x0<midX,y0<midY);

  return client_reshape(c,1);
}

Rule* add_rule(Monitor* mon, char* name, char* role, char* classN, char* classC, Tile* t){
  Rule* r=malloc(sizeof(Rule));
  if(name)r->name=strdup(name);
  else r->name=0;
  if(role)r->role=strdup(role);
  else r->role=0;
  if(classN)r->classN=strdup(classN);
  else r->classN=0;
  if(classC)r->classC=strdup(classC);
  else r->classC=0;

  r->x=r->y=r->w=r->h=RULE_INTNONE;
  r->xr=r->yr=r->wr=r->hr=0;
  r->float_state=RULE_INTNONE;
  r->stack_level=RULE_INTNONE;

  r->target=t;

  r->visible_only=0;
  r->no_steal=0;

  r->next=0;
  if(!mon->rulz){
    mon->rulz=r;
  }else{
    for(Rule* rl=mon->rulz;rl;rl=rl->next){
      if(!rl->next){
	rl->next=r;
	break;
      }
    }
  }
  return r;
}

void remove_rule(Monitor* mon, Rule* r){
  if(mon->rulz==r){
    mon->rulz=r->next;
  }else{
    for(Rule* rl=mon->rulz;rl;rl=rl->next){
      if(rl->next==r){
	rl->next=r->next;
	break;
      }
    }
  }
  destroy_rule(r);
}

void destroy_rule(Rule* r){
  if(!r)return;
  if(r->name)free(r->name);
  if(r->role)free(r->role);
  if(r->classN)free(r->classN);
  if(r->classC)free(r->classC);
  free(r);
}

int tile_is_free(Tile* t){
  if(!t)return 0;
  if(!t->max_clients)return 1;
  if(t->clientcount<t->max_clients)return 1;
  return 0;
}

Tile* tile_cycle(Tile* t){
  if(t->next)return t->next;
  return t->layer->firsttile;
}

Layer* layer_cycle(Layer* l){
  if(l->snext)return l->snext;
  return l->mon->sfirstlayer;
}

Client* client_recycle(Client* c, int mode){
  if(!c)return 0;
  Client* cl=c;
  do{
    cl=client_cycle(cl,mode);
    if(!cl || cl==c)return c;
  }while(!cl->visible || cl->stacktype==StckDesktop || cl->stacktype==StckDock || cl->tile->skip_focus);
  return cl;
}

Client* client_cycle(Client* c, int mode){
  if(!c)return c;
  switch(mode){
    case CycleAll:{//next within all visible layers
      if(c->fnext)return c->fnext;
      return c->mon->firstfocus;
    }
    case CycleTiles:{
      Tile* t=c->tile;
      do{
	t=tile_cycle(t);
      }while(t->clientcount==0 && t!=c->tile);
      for(Client* cc=c->mon->firstfocus;cc;cc=cc->fnext){
	if(cc->tile==t)return cc;
      }
      return t->firstclient;//never happens
    }
    case CycleTileFwd:{
      if(c->snext)return c->snext;
      return c->tile->firstclient;
    }
    case CycleTileBwd:{
      if(c->sprev)return c->sprev;
      return c->tile->lastclient;
    }
  }
  return c;
}


Client* new_client(Window w){
  Client* c=malloc(sizeof(Client));
  if(!c)return 0;
  memset(c,0,sizeof(Client));
  c->float_state=FloatFloat;
  c->stacktype=StckNormal;
  c->win=w;
  return c;
}

Tile* new_tile(const char* name, int x, int y, int w, int h){
  Tile* t=malloc(sizeof(Tile));
  if(!t)return 0;
  memset(t,0,sizeof(Tile));
  if(name)strncpy(t->name,name,255);
  t->xa=x;
  t->ya=y;
  t->wa=w;
  t->ha=h;
  t->sticky=StickNormal;
  return t;
}

Tile* new_tile_full(const char* name, int x, int y, int w, int h, float xr, float yr, float wr, float hr){
  Tile* t=malloc(sizeof(Tile));
  if(!t)return 0;
  memset(t,0,sizeof(Tile));
  if(name)strncpy(t->name,name,255);
  t->xa=x;
  t->ya=y;
  t->wa=w;
  t->ha=h;
  t->xr=xr;
  t->yr=yr;
  t->wr=wr;
  t->hr=hr;
  t->sticky=StickNormal;
  return t;
}


Layer* new_layer(const char* name){
  Layer* l=malloc(sizeof(Layer));
  if(!l)return 0;
  memset(l,0,sizeof(Layer));
  if(name)strncpy(l->name,name,255);
  l->stacktype=StckNormal;
  return l;
}

Monitor* new_monitor(int w, int h){
  Monitor* m=malloc(sizeof(Monitor));
  if(!m)return 0;
  memset(m,0,sizeof(Monitor));
  m->mw=w;
  m->mh=h;
  return m;
}

Dsktp* new_desktop(const char* name, const char* layout){
  Dsktp* d=malloc(sizeof(Dsktp));
  if(!d)return 0;
  memset(d->name,0,256);
  if(name)strncpy(d->name,name,255);
  int len=strlen(layout);
  if(len>64)len=64;
  memset(d->layout,'*',64);
  memcpy(d->layout,layout,len);
  d->next=0;
  d->prev=0;
  d->mon=0;
  d->id=-1;
  return d;
}

Client* win_to_client(Monitor* mon, Window w){
  for(Client* c=mon->firstclient;c;c=c->next){
    if(c->win==w)return c;
  }
  return 0;
}

Client* str_to_client(Monitor* mon, const char* name){
  for(Client* c=mon->firstclient;c;c=c->next){
    if(!strcmp(c->name,name))return c;
  }
  return 0;
}

Layer* str_to_layer(Monitor* mon, const char* name){
  for(Layer* l=mon->firstlayer;l;l=l->next){
    if(!strcmp(l->name,name))return l;
  }
  return 0;
}

Tile* str_to_tile(Monitor* mon, const char* name){
  for(Layer* l=mon->firstlayer;l;l=l->next){
    for(Tile* t=l->firsttile;t;t=t->next){
      if(!strcmp(t->name,name))return t;
    }
  }
  return 0;
}

Tile* str_to_ltile(Layer* l, const char* name){
  for(Tile* t=l->firsttile;t;t=t->next){
    if(!strcmp(t->name,name))return t;
  }
  return 0;
}

Tile* id_to_ltile(Layer* l, int id){
  int index=0;
  for(Tile* t=l->firsttile;t;t=t->next){
    if(index==id)return t;
    index++;
  }
  return 0;
}


Layer* id_to_layer(Monitor* mon, int id){
  int index=0;
  for(Layer* l=mon->firstlayer;l;l=l->next){
    if(index==id)return l;
    index++;
  }
  return 0;
}

int layer_to_id(Monitor* mon, Layer* lay){
  int index=0;
  for(Layer* l=mon->firstlayer;l;l=l->next){
    if(l==lay)return index;
    index++;
  }
  return -1;
}

Dsktp* id_to_desktop(Monitor* mon, int id){
  for(Dsktp* d=mon->firstdesktop;d;d=d->next){
    if(d->id==id)return d;
  }
  return 0;
}

Dsktp* str_to_desktop(Monitor* mon, const char* name){
  for(Dsktp* d=mon->firstdesktop;d;d=d->next){
    if(!strcmp(d->name,name))return d;
  }
  return 0;
}


void tree_print(Monitor* m){
  for(Layer* l=m->sfirstlayer;l;l=l->snext){
    say(">LAYER (%s)\n",l->name);
    for(Tile* t=l->firsttile;t;t=t->next){
      say(">>TILE (%s) |%p < %p > %p|\n",t->name,t->prev,t,t->next);
      say("  clientcount=%d/%d\n",t->clientcount,t->max_clients);
      for(Client* c=t->firstclient;c;c=c->snext){
	say(">>CLIENT (%s) |%p < %p > %p|\n",c->name,c->sprev,c,c->snext);
	say("  win=%x\n",c->win);
	say("  tile=%s\n",c->tile->name);
	say("  visible=%d\n",c->visible);
	say("  hints=%d\n",c->hints);
	say("  stack=%d\n",c->stacktype);
	say("  geom=(%d,%d):(%d,%d)\n",c->x,c->y,c->w,c->h);
      }
    }
  }
}

int rect_isok(Rect* r){
  if(r->x2 < r->x1 + MIN_RECT)return 0;
  if(r->y2 < r->y1 + MIN_RECT)return 0;
  return 1;
}

int rect_intersect(Rect* r1, Rect* r2, Rect* r){
  if(r1->x1 > r2->x1){
    r->x1 = r1->x1;
  }else{
    r->x1 = r2->x1;
  }
  if(r1->x2 > r2->x2){
    r->x2 = r2->x2;
  }else{
    r->x2 = r1->x2;
  }
  if(r1->y1 > r2->y1){
    r->y1 = r1->y1;
  }else{
    r->y1 = r2->y1;
  }
  if(r1->y2 > r2->y2){
    r->y2 = r2->y2;
  }else{
    r->y2 = r1->y2;
  }
  return rect_isok(r);
}

int rect_generate(Client* c, Rect* ptr){
  Tile* t=c->tile;
  //ptr[0]: top
  ptr[0].x1=t->x;
  ptr[0].x2=t->x+t->w;
  ptr[0].y1=t->y;
  ptr[0].y2=c->y;
  //ptr[1]: left
  ptr[1].x1=t->x;
  ptr[1].x2=c->x;
  ptr[1].y1=t->y;
  ptr[1].y2=t->y+t->h;
  //ptr[2]: bottom
  ptr[2].x1=t->x;
  ptr[2].x2=t->x+t->w;
  ptr[2].y1=c->y+c->h;
  ptr[2].y2=t->y+t->h;
  //ptr[2]: right
  ptr[3].x1=c->x+c->w;
  ptr[3].x2=t->x+t->w;
  ptr[3].y1=t->y;
  ptr[3].y2=t->y+t->h;
  return 0;
}

int rect_apply(Rect* ptr, Rect* target, Rect* out, int* outcnt, int limit){
  for(int i=0;i<4;i++){
    if(*outcnt >= limit)return 1;
    if(!rect_isok(ptr+i))continue;
    if(rect_intersect(ptr+i,target,out+(*outcnt))){
      (*outcnt)++;
    }
  }
  return 0;
}

int rect_findbest(Client* c, Tile* t){
  //if type is dialog, or if it has no type, then treat it as a dialog
  if(c->owner && (c->hints&MY_HINT_DIALG || !(c->hints&MY_HINT_TYPES)) && c->tile==t){
    c->x=c->owner->x+c->owner->w/2-c->w/2;
    c->y=c->owner->y+c->owner->h/2-c->h/2;
    return 0;
  }


  Rect ptr[4];
  int limit=256;
  Rect buf1[limit];
  int count1=0;
  Rect buf2[limit];
  int count2=0;
  Rect* nowbuffer=buf1;
  Rect* secondbuffer=buf2;
  int* nowcount=&count1;
  int* secondcount=&count2;
  for(Client* cc=t->firstclient;cc;cc=cc->snext){
    if(cc==c)continue;
    //say("found client %s (%s)\n",cc->name,cc->tile->name);
    //generate bounding rects
    rect_generate(cc,ptr);
    //now the important part
    //reset count on current buffer
    (*nowcount)=0;
    //say("previous count %d\n",*secondcount);
    //if empty, just add
    if(!(*secondcount)){
      for(int i=0;i<4;i++){
	if(rect_isok(ptr+i))nowbuffer[(*nowcount)++]=ptr[i];
      }
    }else{
      for(int i=0;i<(*secondcount);i++){//second buffer contains previous
	if(rect_apply(ptr,secondbuffer+i,nowbuffer,nowcount,limit)){
	  say("rectangle limit overflow, returning\n");
	  return 1;
	}
      }
    }
    //say("new count %d\n",*nowcount);
    //swap buffers
    Rect* tmp=nowbuffer;
    nowbuffer=secondbuffer;
    secondbuffer=tmp;
    int* tmpi=nowcount;
    nowcount=secondcount;
    secondcount=tmpi;
  }
  if(!(*secondcount))return 1;
  int bestfit=-t->w*t->h;
  int bestindex=0;
  //second buffer contains the result
  for(int i=0;i<(*secondcount);i++){
    int rw=secondbuffer[i].x2-secondbuffer[i].x1;
    int rh=secondbuffer[i].y2-secondbuffer[i].y1;
    int newfit=0;
    if(rw<c->w)newfit-=c->h*(c->w-rw);
    if(rh<c->h)newfit-=c->w*(c->h-rh);
    if(newfit>bestfit){
      bestindex=i;
      bestfit=newfit;
    }
  }
  c->x=secondbuffer[bestindex].x1;
  c->y=secondbuffer[bestindex].y1;
  return 0;
}
